# -*- coding: utf_8 -*-
"""Pattern Macher."""
import copy

from libsast.core_matcher.helpers import get_rules
from libsast.core_matcher import matchers
from libsast import (
    common,
    exceptions,
)


class PatternMatcher:
    def __init__(self, options: dict) -> None:
        self.matcher = matchers.MatchCommand()
        self.scan_rules = get_rules(options.get('match_rules'))
        self.show_progress = options.get('show_progress')
        exts = options.get('match_extensions')
        if exts:
            self.exts = [ext.lower() for ext in exts]
        else:
            self.exts = []
        self.findings = {}

    def scan(self, paths: list) -> dict:
        """Scan file(s) or directory."""
        if not self.scan_rules:
            return
        self.validate_rules()
        if self.show_progress:
            pbar = common.ProgressBar('Pattern Match', len(paths))
            paths = pbar.progrees_loop(paths)
        for sfile in paths:
            if self.exts:
                if not sfile.suffix.lower() in self.exts:
                    continue
            data = sfile.read_text('utf-8', 'ignore')
            self.pattern_matcher(data, sfile.as_posix())
        return self.findings

    def validate_rules(self):
        """Validate Rules before scanning."""
        for rule in self.scan_rules:
            if not isinstance(rule, dict):
                raise exceptions.InvalidRuleFormatException(
                    'Pattern Matcher Rule format is invalid.')
            if not rule.get('type'):
                raise exceptions.TypeKeyMissingError(
                    'The rule is missing the key \'type\'')
            if not rule.get('pattern'):
                raise exceptions.PatternKeyMissingError(
                    'The rule is missing the key \'pattern\'')
            all_mts = [m for m in dir(matchers) if m.startswith('R')]
            pattern_name = rule['type']
            if pattern_name not in all_mts:
                supported = ', '.join(all_mts)
                raise exceptions.MatcherNotFoundException(
                    f'Matcher \'{pattern_name}\' is not supported.'
                    f' Available matchers are {supported}',
                )

    def pattern_matcher(self, data, file_path):
        """Static Analysis Pattern Matcher."""
        try:
            for rule in self.scan_rules:
                case = rule.get('input_case')
                if case == 'lower':
                    fmt_data = data.lower()
                elif case == 'upper':
                    fmt_data = data.upper()
                else:
                    fmt_data = data
                matches = self.matcher._find_match(
                    rule['type'],
                    fmt_data,
                    rule)
                if matches:
                    self.add_finding(file_path, rule, matches)
        except Exception:
            raise exceptions.RuleProcessingException('Rule processing error.')

    def add_finding(self, file_path, rule, matches):
        """Add Code Analysis Findings."""
        for match in matches:
            crule = copy.deepcopy(rule)
            file_details = {
                'file_path': file_path,
                'match_string': match[0],
                'match_position': match[1],
                'match_lines': match[2],
            }
            if rule['id'] in self.findings:
                self.findings[rule['id']]['files'].append(file_details)
            else:
                self.findings[rule['id']] = {
                    'files': [file_details],
                    'metadata': crule,
                }



# -*- coding: utf_8 -*-
"""Semantic Grep Core."""
from libsast.core_sgrep.helpers import invoke_semgrep
from libsast import common


class SemanticGrep:
    def __init__(self, options: dict) -> None:
        self.scan_rules = options.get('sgrep_rules')
        self.show_progress = options.get('show_progress')
        exts = options.get('sgrep_extensions')
        if exts:
            self.exts = [ext.lower() for ext in exts]
        else:
            self.exts = []
        self.findings = {
            'matches': {},
            'errors': [],
        }

    def scan(self, paths: list) -> dict:
        """Do sgrep scan."""
        if self.exts:
            filtered = []
            for sfile in paths:
                if sfile.suffix.lower() in self.exts:
                    filtered.append(sfile)
            if filtered:
                paths = filtered
        if self.show_progress:
            pbar = common.ProgressBar('Semantic Grep', len(paths))
            sgrep_out = pbar.progress_function(
                invoke_semgrep,
                (paths, self.scan_rules))
        else:
            sgrep_out = invoke_semgrep(paths, self.scan_rules)
        self.format_output(sgrep_out)
        return self.findings

    def format_output(self, results):
        """Format sgrep results."""
        self.findings['errors'] = results['errors']
        smatches = self.findings['matches']
        for find in results['results']:
            file_details = {
                'file_path': find['path'],
                'match_position': (find['start']['col'], find['end']['col']),
                'match_lines': (find['start']['line'], find['end']['line']),
                'match_string': find['extra']['lines'],
            }
            rule_id = find['check_id'].rsplit('.', 1)[1]
            if rule_id in smatches:
                smatches[rule_id]['files'].append(file_details)
            else:
                metadata = find['extra']['metadata']
                metadata['description'] = find['extra']['message']
                metadata['severity'] = find['extra']['severity']
                smatches[rule_id] = {
                    'files': [file_details],
                    'metadata': metadata,
                }



# -*- coding: utf_8 -*-
"""Common Helpers."""
import sys
from threading import Thread

from libsast.exceptions import (
    YamlRuleLoadException,
    YamlRuleParseError,
)

import yaml


class ProgressBar:
    def __init__(self, prefix, expected_time, size=60, output=sys.stderr):
        self.prefix = prefix
        self.expected_time = expected_time
        self.size = size
        self.output = output

    def progress_print(self, index):
        """Print progress bar."""
        prog_length = int(self.size * index / self.expected_time)
        prog = '█' * prog_length
        self.output.write(f'- {self.prefix} {prog} {index}\r')
        self.output.flush()

    def progrees_loop(self, iterator):
        """Show progress for loop."""
        self.progress_print(0)
        for index, item in enumerate(iterator):
            yield item
            self.progress_print(index + 1)
        self.output.write('\n')
        self.output.flush()

    def progress_function(self, function, args=None, kwargs=None):
        """Show progress for function."""
        ret = [None]
        index = 0
        # Hack determined by size of rule files
        self.expected_time = self.expected_time * 6

        def myrunner(function, ret, *args, **kwargs):
            ret[0] = function(*args, **kwargs)
        self.progress_print(0)
        thread = Thread(
            target=myrunner,
            args=(function, ret) + tuple(args),
            kwargs=kwargs)
        thread.start()
        while thread.is_alive():
            thread.join(timeout=0.2)
            index += 1
            self.progress_print(index)
        self.output.write('\n')
        self.output.flush()
        return ret[0]


def read_yaml(file_obj, text=False):
    try:
        if text:
            return yaml.safe_load(file_obj)
        return yaml.safe_load(file_obj.read_text('utf-8', 'ignore'))
    except yaml.YAMLError as exp:
        raise YamlRuleParseError(
            f'YAML Parse Error: {repr(exp)}')
    except Exception as gen:
        raise YamlRuleLoadException(
            f'Failed to load YAML file: {repr(gen)}')



# -*- coding: utf_8 -*-
"""Helper Functions."""
from pathlib import Path

from libsast.exceptions import (
    InvalidRuleError,
    MissingRuleError,
    RuleDownloadException,
)
from libsast.common import read_yaml
from libsast.standards import get_mapping

import requests


def download_rule(url):
    """Download Pattern File."""
    try:
        with requests.get(url, allow_redirects=True) as r:
            r.raise_for_status()
            return r.text
    except requests.exceptions.RequestException:
        raise RuleDownloadException(f'Failed to download from: {url}')


def get_rules(rule_loc):  # noqa: R701
    """Get pattern matcher rules."""
    if not rule_loc:
        raise MissingRuleError('Rule location is missing.')
    if rule_loc.startswith(('http://', 'https://')):
        pat = download_rule(rule_loc)
        if not pat:
            return
        rules = read_yaml(pat, True)
        if not rules:
            return
        return get_mapping(rules)
    rule = Path(rule_loc)
    if rule.is_file() and rule.exists():
        rules = read_yaml(rule)
        if not rules:
            return
        return get_mapping(rules)
    elif rule.is_dir() and rule.exists():
        patterns = []
        for yfile in rule.glob('**/*.yaml'):
            rules = read_yaml(yfile)
            if rules:
                rules = get_mapping(rules)
                patterns.extend(rules)
        return patterns
    else:
        raise InvalidRuleError('This path is invalid')



# -*- coding: utf_8 -*-
"""Semantic Grep Helpers."""
import json
import platform
import multiprocessing
from io import StringIO


def invoke_semgrep(paths, scan_rules, **kwargs):
    """Call Semgrep."""
    if platform.system() == 'Windows':
        return None
    from semgrep import semgrep_main, util
    from semgrep.constants import OutputFormat
    from semgrep.output import OutputHandler, OutputSettings
    try:
        cpu_count = multiprocessing.cpu_count()
    except NotImplementedError:
        cpu_count = 1  # CPU count is not implemented on Windows
    util.set_flags(False, True, False)  # Verbose, Quiet, Force_color

    io_capture = StringIO()
    output_handler = OutputHandler(
        OutputSettings(
            output_format=OutputFormat.JSON,
            output_destination=None,
            error_on_findings=False,
            strict=False,
        ),
        stdout=io_capture,
    )
    semgrep_main.main(
        output_handler=output_handler,
        target=[pt.as_posix() for pt in paths],
        jobs=cpu_count,
        pattern=None,
        lang=None,
        config=scan_rules,
        timeout=5,
        **kwargs,
    )
    output_handler.close()
    return json.loads(io_capture.getvalue())



# -*- coding: utf_8 -*-
"""The libsast Exceptions."""


class LibsastException(Exception):
    """Base class for all exceptions thrown by libsast."""

    def __init__(self, message=None):
        super().__init__(message)


class InvalidPathError(LibsastException):
    """Invalid Path Supplied to libsast."""

    pass


class YamlRuleParseError(LibsastException):
    """Failed to parse YAML rule."""

    pass


class YamlRuleLoadException(LibsastException):
    """Failed to load YAML rule file."""

    pass


class MissingRuleError(LibsastException):
    """Rule not provided."""

    pass


class InvalidRuleError(LibsastException):
    """No rule directory, file or url specified."""

    pass


class TypeKeyMissingError(LibsastException):
    """Pattern Matcher rule does not have the key 'type'."""

    pass


class InvalidRuleFormatException(LibsastException):
    """Pattern Matcher rule file is invalid."""

    pass


class PatternKeyMissingError(LibsastException):
    """Pattern Matcher rule does not have the key 'pattern'."""

    pass


class RuleDownloadException(LibsastException):
    """Failed to download rule."""

    pass


class RuleProcessingException(LibsastException):
    """Failed to download rule."""

    pass


class MatcherNotFoundException(LibsastException):
    """Pattern Matcher not found."""

    pass



# !/usr/bin/env python
# -*- coding: utf_8 -*-
"""libsast cli."""
import argparse
import json
import sys

from libsast import __version__
from libsast.scanner import Scanner


def output(out, scan_results):
    """Output."""
    if out:
        with open(out, 'w') as outfile:
            json.dump(scan_results,
                      outfile,
                      sort_keys=True,
                      indent=2,
                      separators=(',', ': '))
    else:
        if scan_results:
            print(json.dumps(scan_results,
                             sort_keys=True,
                             indent=2,
                             separators=(',', ': ')))
    if scan_results:
        sgrep_out = scan_results.get('semantic_grep', {}).get('matches')
        matcher_out = scan_results.get('pattern_matcher')
        if sgrep_out or matcher_out:
            sys.exit(1)
    sys.exit(0)


def main():
    """Main CLI."""
    parser = argparse.ArgumentParser()
    parser.add_argument('path',
                        nargs='*',
                        help=('Path can be file(s) or '
                              'directories'))
    parser.add_argument('-o', '--output',
                        help='Output filename to save JSON report.',
                        required=False)
    parser.add_argument('-p', '--pattern-file',
                        help='YAML pattern file, directory or url',
                        required=False)
    parser.add_argument('-s', '--sgrep-pattern-file',
                        help='sgrep rules directory',
                        required=False)
    parser.add_argument('--sgrep-file-extensions',
                        nargs='+',
                        help=('File extensions that should be scanned'
                              ' with sgrep'),
                        required=False)
    parser.add_argument('--file-extensions',
                        nargs='+',
                        help=('File extensions that should be scanned'
                              ' with pattern matcher'),
                        required=False)
    parser.add_argument('--ignore-filenames',
                        nargs='+',
                        help='File name(s) to ignore',
                        required=False)
    parser.add_argument('--ignore-extensions',
                        nargs='+',
                        help='File extension(s) to ignore in lower case',
                        required=False)
    parser.add_argument('--ignore-paths',
                        nargs='+',
                        help='Path(s) to ignore',
                        required=False)
    parser.add_argument('--show-progress',
                        help='Show scan progress',
                        required=False,
                        action='store_true')
    parser.add_argument('-v', '--version',
                        help='Show libsast version',
                        required=False,
                        action='store_true')
    args = parser.parse_args()
    if args.path and (args.pattern_file or args.sgrep_pattern_file):
        options = {
            'sgrep_rules': args.sgrep_pattern_file,
            'sgrep_extensions': args.sgrep_file_extensions,
            'match_rules': args.pattern_file,
            'match_extensions': args.file_extensions,
            'ignore_filenames': args.ignore_filenames,
            'ignore_extensions': args.ignore_extensions,
            'ignore_paths': args.ignore_paths,
            'show_progress': args.show_progress,
        }
        result = Scanner(options, args.path).scan()
        output(args.output, result)
    elif args.version:
        print(f'libsast v{__version__}')
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



# -*- coding: utf_8 -*-
"""The libsast Scanner module."""
from pathlib import Path

from libsast.core_matcher.pattern_matcher import (
    PatternMatcher,
)
from libsast.core_sgrep.semantic_sgrep import (
    SemanticGrep,
)
from libsast.exceptions import (
    InvalidPathError,
)


class Scanner:
    def __init__(self, options: dict, paths: list) -> None:
        if options:
            self.options = options
        else:
            self.options = {
                'sgrep_rules': None,
                'sgrep_extensions': None,
                'match_rules': None,
                'match_extensions': None,
                'ignore_filenames': None,
                'ignore_extensions': None,
                'ignore_paths': None,
                'show_progress': False,
            }
        if options.get('ignore_extensions'):
            self.ignore_extensions = options.get('ignore_extensions')
        else:
            self.ignore_extensions = []
        if options.get('ignore_filenames'):
            self.ignore_filenames = options.get('ignore_filenames')
        else:
            self.ignore_filenames = []
        if options.get('ignore_paths'):
            self.ignore_paths = options.get('ignore_paths')
        else:
            self.ignore_paths = []
        self.paths = paths

    def scan(self) -> dict:
        """Start Scan."""
        results = {}
        valid_paths = self.get_scan_files(self.paths)
        if not valid_paths:
            return
        if self.options.get('match_rules'):
            results['pattern_matcher'] = PatternMatcher(
                self.options).scan(valid_paths)
        if self.options.get('sgrep_rules'):
            results['semantic_grep'] = SemanticGrep(
                self.options).scan(valid_paths)
        return results

    def get_scan_files(self, paths):
        """Get files valid for scanning."""
        if not isinstance(paths, list):
            raise InvalidPathError('Path should be a list')
        all_files = set()
        for path in paths:
            pobj = Path(path)
            if pobj.is_dir():
                for pfile in pobj.rglob('*'):
                    if self.validate_file(pfile):
                        all_files.add(pfile)
            else:
                if self.validate_file(pobj):
                    all_files.add(pobj)
        return all_files

    def validate_file(self, path):
        """Check if we should scan the file."""
        ignore_paths = any(pp in path.as_posix() for pp in self.ignore_paths)
        ignore_files = path.name in self.ignore_filenames
        ignore_exts = path.suffix.lower() in self.ignore_extensions
        if (ignore_paths or ignore_files or ignore_exts):
            return False
        if not path.exists() or not path.is_file():
            return False
        return True



# -*- coding: utf_8 -*-
"""Alogrithms for Matching Pattern."""
import re
from abc import ABC, abstractclassmethod


def get_match_lines(content, pos):
    """Get Match lines from position."""
    start_line = 0
    filepos = 0
    skip = False
    for idx, line in enumerate(content.split('\n'), 1):
        filepos += len(line) + 1
        if filepos >= pos[0] and filepos >= pos[1] and not skip:
            # Match is on the same line
            return (idx, idx)
        elif filepos >= pos[0] and not skip:
            # Multiline match, find start line
            skip = True
            start_line = idx
        if filepos >= pos[1] and skip:
            # Multiline march, find end line
            return (start_line, idx)


class MatchCommand:

    def __init__(self):
        self.patterns = {}

    def _find_match(self, pattern_name, content, rule):
        if pattern_name not in self.patterns:
            pattern_class = globals()[pattern_name]
            self.patterns[pattern_name] = pattern_class()
        return self.patterns[pattern_name]._perform_search(content, rule)


class MatchStrategy(ABC):

    @abstractclassmethod
    def _perform_search(self, content, rule):
        """Search for instance of match in content."""


class Regex(MatchStrategy):
    def _perform_search(self, content, rule):
        matches = set()
        for match in re.compile(rule['pattern']).finditer(content):
            if match.group():
                match_pos = match.span()
                match_lines = get_match_lines(content, match_pos)
                matches.add((match.group(), match_pos, match_lines))
        return matches


class RegexAnd(MatchStrategy):
    def _perform_search(self, content, rule):
        if isinstance(rule['pattern'], str):
            return Regex().perform_search(content, rule)
        matches = set()
        for regex in rule['pattern']:
            empty_iter_detected = True
            for match in re.compile(regex).finditer(content):
                empty_iter_detected = False
                match_pos = match.span()
                match_lines = get_match_lines(content, match_pos)
                matches.add((match.group(), match_pos, match_lines))
            if empty_iter_detected:
                return False
        return matches


class RegexOr(MatchStrategy):
    def _perform_search(self, content, rule):
        if isinstance(rule['pattern'], str):
            return Regex().perform_search(content, rule)
        matches = set()
        for regex in rule['pattern']:
            for match in re.compile(regex).finditer(content):
                if match.group():
                    match_pos = match.span()
                    match_lines = get_match_lines(content, match_pos)
                    matches.add((match.group(), match_pos, match_lines))
        return matches


class RegexAndNot(MatchStrategy):
    def _perform_search(self, content, rule):
        matches = set()
        regex_present = re.compile(rule['pattern'][0]).finditer(content)
        regex_not = re.compile(rule['pattern'][1]).finditer(content)
        for match in regex_not:
            if match.group():
                return False
        for match in regex_present:
            if match.group():
                match_pos = match.span()
                match_lines = get_match_lines(content, match_pos)
                matches.add((match.group(), match_pos, match_lines))
        return matches


class RegexAndOr(MatchStrategy):
    def _perform_search(self, content, rule):
        matches = set()
        or_matches = set()
        or_list = rule['pattern'][1]
        break_parent_loop = False
        for regex in or_list:
            for match in re.compile(regex).finditer(content):
                if match.group():
                    match_pos = match.span()
                    match_lines = get_match_lines(content, match_pos)
                    or_matches.add((match.group(), match_pos, match_lines))
                    break_parent_loop = True
                    break
            if break_parent_loop:
                break
        for match in re.compile(rule['pattern'][0]).finditer(content):
            if match.group():
                match_pos = match.span()
                match_lines = get_match_lines(content, match_pos)
                matches.add((match.group(), match_pos, match_lines))
        if not (matches and or_matches):
            return False
        matches.update(or_matches)
        return matches



#!/usr/bin/env python
# -*- coding: utf_8 -*-
from .core_matcher.pattern_matcher import PatternMatcher
from .core_sgrep.semantic_sgrep import SemanticGrep
from .scanner import Scanner

__title__ = 'libsast'
__authors__ = 'Ajin Abraham'
__copyright__ = 'Copyright 2020 Ajin Abraham, OpenSecurity'
__version__ = '1.1.2'
__version_info__ = tuple(int(i) for i in __version__.split('.'))
__all__ = [
    'Scanner',
    'PatternMatcher',
    'SemanticGrep',
    '__title__',
    '__authors__',
    '__copyright__',
    '__version__',
    '__version_info__',
]



# -*- coding: utf_8 -*-
"""Standards Matcher."""
from pathlib import Path

from libsast.common import (
    read_yaml,
)


def read_standards(rule_loc):
    """Get mappings."""
    if rule_loc.is_dir() and rule_loc.exists():
        patterns = {}
        for yfile in rule_loc.glob('**/*.yaml'):
            rule = read_yaml(yfile)
            if rule:
                patterns.update(rule)
        return patterns
    return None


def get_standards():
    """Get inbuilt mappings."""
    stds_loc = Path(__file__).parents[0] / 'standards'
    return read_standards(stds_loc)


def get_mapping(rules):
    """Process rule mappings."""
    std_map = get_standards()
    new_rules = []
    for rule in rules:
        for key in rule.keys():
            if key in std_map.keys():
                rule[key] = std_map[key].get(rule[key], rule[key])
        new_rules.append(rule)
    return new_rules
